﻿//#define FIG1
//#define FIG34

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Drawing.Drawing2D;

namespace howto_hexagonal_grid_rotated
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        // The width of a hexagon.
        private const float HexWidth = 50;

        // Selected hexagons.
        private List<PointF> Hexagons = new List<PointF>();

#if FIG34
        // The selected search rectangle.
        // Used to draw Figures 3 and 4.
        private List<RectangleF> TestRects = new List<RectangleF>();
#endif

        // Redraw the grid.
        private void picGrid_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;

            // Draw the selected hexagons.
            foreach (PointF point in Hexagons)
            {
                e.Graphics.FillPolygon(Brushes.LightBlue,
                    HexToPoints(HexWidth, point.X, point.Y));
            }

            // Draw the grid.
            DrawHexGrid(e.Graphics, Pens.Black,
                0, picGrid.ClientSize.Width,
                0, picGrid.ClientSize.Height,
                HexWidth);

#if FIG34
            // Draw the selected rectangles for Figures 3 and 4.
            using (Pen pen = new Pen(Color.Red, 3))
            {
                pen.DashStyle = DashStyle.Dash;
                foreach (RectangleF rect in TestRects)
                {
                    e.Graphics.DrawRectangle(pen, Rectangle.Round(rect));
                }
            }
#endif
        }

        // Draw a hexagonal grid for the indicated area.
        // (You might be able to draw the hexagons without
        // drawing any duplicate edges, but this is a lot easier.)
        private void DrawHexGrid(Graphics gr, Pen pen,
            float xmin, float xmax, float ymin, float ymax,
            float height)
        {
            // Loop until a hexagon won't fit.
            for (int row = 0; ; row++)
            {
                // Get the points for the row's first hexagon.
                PointF[] points = HexToPoints(height, row, 0);

                // If it doesn't fit, we're done.
                if (points[3].Y > ymax) break;

                // Draw the row.
                for (int col = 0; ; col++)
                {
                    // Get the points for the row's next hexagon.
                    points = HexToPoints(height, row, col);

                    // If it doesn't fit horizontally,
                    // we're done with this row.
                    if (points[1].X > xmax) break;

                    // If it fits vertically, draw it.
                    if (points[3].Y <= ymax)
                    {
                        gr.DrawPolygon(pen, points);

#if FIG1
                        // Label the hexagon (for Figure 1).
                        using (StringFormat sf = new StringFormat())
                        {
                            sf.Alignment = StringAlignment.Center;
                            sf.LineAlignment = StringAlignment.Center;
                            float x = (points[0].X + points[3].X) / 2;
                            float y = (points[1].Y + points[4].Y) / 2;
                            string label = "(" + row.ToString() + ", " +
                                col.ToString() + ")";
                            gr.DrawString(label, this.Font,
                                Brushes.Black, x, y, sf);
                        }
#endif
                    }
                }
            }
        }

        private void picGrid_Resize(object sender, EventArgs e)
        {
            picGrid.Refresh();
        }

        // Display the row and column under the mouse.
        private void picGrid_MouseMove(object sender, MouseEventArgs e)
        {
            int row, col;
            PointToHex(e.X, e.Y, HexWidth, out row, out col);
            this.Text = "(" + row + ", " + col + ")";
        }

        // Add the clicked hexagon to the Hexagons list.
        private void picGrid_MouseClick(object sender, MouseEventArgs e)
        {
            int row, col;
            PointToHex(e.X, e.Y, HexWidth, out row, out col);
            Hexagons.Add(new PointF(row, col));

#if FIG34
            // Used to draw Figures 3 and 4.
            PointF[] points = HexToPoints(HexWidth, row, col);
            TestRects.Add(new RectangleF(
                points[4].X, points[0].Y,
                points[1].X - points[4].X,
                0.75f * (points[3].Y - points[0].Y))
            );
#endif

            picGrid.Refresh();
        }

        // Return the height of a hexagon.
        private float HexHeight(float height)
        {
            return (float)(4 * (height / 2 / Math.Sqrt(3)));
        }

        // Return the row and column of the hexagon at this point.
        private void PointToHex(float x, float y, float width,
            out int row, out int col)
        {
            // Find the test rectangle containing the point.
            float height = HexHeight(width);
            row = (int)(y / (height * 0.75f));

            if (row % 2 == 0)
                col = (int)Math.Floor(x / width);
            else
                col = (int)Math.Floor((x - width / 2) / width);

            // Find the test area.
            float testx = col * width;
            float testy = row * height * 0.75f;
            if (row % 2 != 0) testx += width / 2;

            // See if the point is left or
            // right of the test hexagon above.
            bool is_left = false, is_right = false;
            float dy = y - testy;
            if (dy < height / 4)
            {
                float dx = x - (testx + width / 2);
                if (dy < 0.001)
                {
                    // The point is on the top edge of the test rectangle.
                    if (dy < 0) is_left = true;
                    if (dy > 0) is_right = true;
                }
                else if (dx < 0)
                {
                    // See if the point is left of the test hexagon.
                    if (-dx / dy > Math.Sqrt(3)) is_left = true;
                }
                else
                {
                    // See if the point is right of the test hexagon.
                    if (dx / dy > Math.Sqrt(3)) is_right = true;
                }
            }

            // Adjust the row and column if necessary.
            if (is_left)
            {
                if (row % 2 == 0) col--;
                row--;
            }
            else if (is_right)
            {
                if (row % 2 != 0) col++;
                row--;
            }
        }

        // Return the points that define the indicated hexagon.
        private PointF[] HexToPoints(float width, float row, float col)
        {
            // Start with the top center of the first hexagon.
            float height = HexHeight(width);
            float x = width / 2;
            float y = 0;

            // Move right the required number of rows.
            x += col * width;

            // If the row is odd, move right half a hex more.
            if (row % 2 != 0) x += width / 2;

            // Move over for the row number.
            y += row * (height * 0.75f);

            // Generate the points.
            return new PointF[]
            {
                new PointF(x, y),
                new PointF(x + width / 2, y + height * 0.25f),
                new PointF(x + width / 2, y + height * 0.75f),
                new PointF(x, y + height),
                new PointF(x - width / 2, y + height * 0.75f),
                new PointF(x - width / 2, y + height * 0.25f),
            };
        }
    }
}
